@@ -38,7 +38,10 @@ class LazyGroupDict(Mapping):
         """The Document object, retrieved from provided path
         Note: Path can be both local and remote
         """
-        return self.groups[group_name]
+        if group_name in self.groups:
+            return self.groups[group_name]
+        else:
+            raise KeyError
 
     def __iter__(self):
         ...
