@@ -4,17 +4,19 @@ as well as the OCR data and page images of the document"""
 import copy
 import itertools
 import logging
+import multiprocessing
 import os
 from collections import OrderedDict
 from typing import Dict, List, Tuple, Any, Optional, Callable
 
 import fitz
 import pandas as pd
-from cloudstorageio import CloudInterface
 from fitz.utils import getColor, getColorList
 
+from pycognaize.login import Login
 from pycognaize.common.enums import IqDocumentKeysEnum, FieldTypeEnum
 from pycognaize.common.field_collection import FieldCollection
+from pycognaize.common.utils import cloud_interface_login
 from pycognaize.document.field import FieldMapping, TableField
 from pycognaize.document.field.field import Field
 from pycognaize.document.page import Page
@@ -32,6 +34,7 @@ class Document:
                  output_fields: 'FieldCollection[str, List[Field]]',
                  pages: Dict[int, Page],
                  metadata: Dict[str, Any]):
+        self._login_instance = Login()
         self._metadata = metadata
         self._pages: Dict[int, Page] = pages
         self._x: FieldCollection[str, List[Field]] = input_fields
@@ -152,7 +155,8 @@ class Document:
     def get_tied_fields(self, tag: ExtractionTag,
                         field_type: str = FieldTypeEnum.BOTH.value,
                         threshold: float = 0.5,
-                        pn_filter: Callable = lambda x: True):
+                        pn_filter: Callable = lambda x: True
+                        ) -> Dict[str, List[Field]]:
         """Given an `ExtractionTag`, return all the fields that contain
            tags in the same physical location.
 
@@ -162,9 +166,10 @@ class Document:
             location
         :param pn_filter: If provided, only fields with
             names passing the filter will be considered
-        :return: List of `Field` objects
+        :return: Dictionary where key is pname
+            and value is List of `Field` objects
         """
-        all_tied_fields = []
+        all_tied_fields: Dict[str, List[Field]] = OrderedDict()
         if field_type == FieldTypeEnum.INPUT_FIELD.value:
             scopes = (self.x,)
         elif field_type == FieldTypeEnum.OUTPUT_FIELD.value:
@@ -180,21 +185,21 @@ class Document:
             for pname, fields in scope.items():
                 if not pn_filter(pname):
                     continue
-                tied_fields = [field for field in fields
+                tied_fields = {field for field in fields
                                if not isinstance(field, TableField)
                                for field_tag in field.tags
                                if isinstance(field_tag, ExtractionTag)
                                and (tag & field_tag)
-                               / min({tag.area, field_tag.area}) >= threshold]
+                               / min({tag.area, field_tag.area}) >= threshold}
                 if tied_fields:
-                    all_tied_fields.extend(tied_fields)
+                    all_tied_fields[pname] = list(tied_fields)
         return all_tied_fields
 
     def get_tied_tags(self, tag: ExtractionTag,
                       field_type: str = FieldTypeEnum.BOTH.value,
                       threshold: float = 0.9,
                       pn_filter: Callable = lambda x: True
-                      ) -> List[ExtractionTag]:
+                      ) -> Dict[str, List[ExtractionTag]]:
         """Given a single tag, return all other tags in the document that are
             in the same physical location in the document
 
@@ -204,9 +209,10 @@ class Document:
             location
         :param pn_filter: If provided, only tags that are in fields
             with names passing the filter will be considered
-        :return: List of `ExtractionTag` objects
+        :return: Dictionary where key is pname
+            and value is List of `ExtractionTag` objects
         """
-        all_tied_tags = []
+        all_tied_tags: Dict[str, List[ExtractionTag]] = OrderedDict()
         if field_type == FieldTypeEnum.INPUT_FIELD.value:
             scopes = (self.x,)
         elif field_type == FieldTypeEnum.OUTPUT_FIELD.value:
@@ -222,31 +228,33 @@ class Document:
             for pname, fields in scope.items():
                 if not pn_filter(pname):
                     continue
-                tied_tags = [field_tag for field in fields
+                tied_tags = {field_tag for field in fields
                              if not isinstance(field, TableField)
                              for field_tag in field.tags
                              if isinstance(field_tag, ExtractionTag)
-                             if tag.iou(field_tag) >= threshold]
+                             if tag.iou(field_tag) >= threshold}
                 if tied_tags:
-                    all_tied_tags.extend(tied_tags)
+                    all_tied_tags[pname] = list(tied_tags)
         return all_tied_tags
 
     def get_first_tied_field(self, tag: ExtractionTag,
                              pn_filter: Callable = lambda x: True
-                             ) -> Optional[Field]:
+                             ) -> Tuple[str, Field]:
         """Return the first field that is in the same location as the given tag
 
         :param tag: Input `ExtractionTag`
         :param pn_filter: If provided, only fields with
             names passing the filter will be considered
-        :return: If match found, return the matching `Field`,
-            otherwise return `None`
+        :return: If match found, return Tuple of the matching pname
+            and `Field`, otherwise return `None`
         """
         res = None
         fields = self.get_tied_fields(tag=tag,
                                       pn_filter=pn_filter)
-
-        res = fields[0] if fields else res
+        if fields:
+            pname = list(fields)[0]
+            first_tied_field = fields[pname][0]
+            res = (pname, first_tied_field) if first_tied_field else res
         return res
 
     def get_first_tied_field_value(self, tag: ExtractionTag,
@@ -263,30 +271,35 @@ class Document:
         if isinstance(tag, float):
             val = ''
         else:
-            matching_field = self.get_first_tied_field(
+            tied_field = self.get_first_tied_field(
                 tag=tag, pn_filter=pn_filter)
-            if matching_field is None:
+            if tied_field is None:
                 val = tag.raw_value
             else:
+                pname, matching_field = self.get_first_tied_field(
+                    tag=tag, pn_filter=pn_filter)
                 # noinspection PyUnresolvedReferences
                 val = matching_field.value
         return val
 
     def get_first_tied_tag(self, tag: ExtractionTag,
                            pn_filter: Callable = lambda x: True
-                           ) -> Optional[ExtractionTag]:
+                           ) -> Tuple[str, ExtractionTag]:
         """Return the first tag that is in the same location as the given tag
 
         :param tag: Input `ExtractionTag`
         :param pn_filter: If provided, only tags that
             are in fields with names passing the filter will be considered
-        :return: If match found, return the matching `ExtractionTag`,
-            otherwise return `None`
+        :return: If match found, return Tuple of the matching pname
+            and `ExtractionTag`, otherwise return `None`
         """
         res = None
         tags = self.get_tied_tags(tag=tag,
                                   pn_filter=pn_filter)
-        res = tags[0] if tags else res
+        if tags:
+            pname = list(tags)[0]
+            first_tied_tag = tags[pname][0]
+            res = (pname, first_tied_tag) if first_tied_tag else res
         return res
 
     def get_first_tied_tag_value(self, tag: ExtractionTag,
@@ -300,12 +313,13 @@ class Document:
             fields with names passing the filter will be considered
         :return:
         """
-        matching_tag = self.get_first_tied_tag(tag=tag,
-                                               pn_filter=pn_filter)
-        if matching_tag is None:
+        tied_tag = self.get_first_tied_tag(tag=tag,
+                                           pn_filter=pn_filter)
+        if tied_tag is None:
             val = tag.raw_value
         else:
-            val = matching_tag.value
+            pname, matching_tag = tied_tag
+            val = matching_tag.raw_value
         return val
 
     def get_df_with_tied_field_values(self, table_tag: TableTag,
@@ -327,6 +341,34 @@ class Document:
                 x,
                 pn_filter=pn_filter))
 
+    def load_page_images(self, page_filter: Callable = lambda x: True) -> None:
+        """Get all images of the pages in the document
+         (Using multiprocessing)"""
+        global _get_page
+
+        def _get_page(page, filter_pages: Callable = page_filter):
+            if filter_pages(page):
+                _ = page.image_bytes
+            return page
+
+        pool = multiprocessing.Pool(min(multiprocessing.cpu_count() * 2, 16))
+        pages = pool.map(_get_page, self.pages.values())
+        self._pages = OrderedDict({page.page_number: page for page in pages})
+
+    def load_page_ocr(self, page_filter: Callable = lambda x: True) -> None:
+        """Get all OCR of the pages in the document
+           (Using multiprocessing)"""
+        global _get_page
+
+        def _get_page(page, filter_pages: Callable = page_filter):
+            if filter_pages(page):
+                _ = page.ocr
+            return page
+
+        pool = multiprocessing.Pool(min(multiprocessing.cpu_count() * 2, 16))
+        pages = pool.map(_get_page, self.pages.values())
+        self._pages = OrderedDict({page.page_number: page for page in pages})
+
     def to_dict(self) -> dict:
         """Converts Document object to dict"""
         input_fields = OrderedDict(
@@ -341,10 +383,12 @@ class Document:
         return data
 
     @classmethod
-    def from_dict(cls, raw: dict, data_path: str) -> 'Document':
+    def from_dict(cls, raw: dict,
+                  data_path: str) -> 'Document':
         """Document object created from data of dict
         :param raw: document dictionary
         :param data_path: path to the documents OCR and page images
+        :param login_instance: login instance of pycognaize
         """
         if not isinstance(raw, dict):
             raise TypeError(
@@ -369,7 +413,8 @@ class Document:
                 ].value.construct_from_raw(raw=field, pages=pages)
                 for field in fields]
              for name, fields in raw['output_fields'].items()})
-        return cls(input_fields=input_fields, output_fields=output_fields,
+        return cls(input_fields=input_fields,
+                   output_fields=output_fields,
                    pages=pages, metadata=metadata)
 
     def _collect_all_tags_for_fields(self,
@@ -424,7 +469,7 @@ class Document:
         :return: bytes object of the pdf
         """
 
-        ci = CloudInterface()
+        ci = cloud_interface_login(self._login_instance)
         pdf_path = os.path.join(self.pages[1].path, self.document_src) + '.pdf'
 
         with ci.open(pdf_path, 'rb') as f:
